(**************************************************************************************)
(*  'tils Lib - Librairie de composants divers                                        *)
(*  Copyright (C) 2003  U. Van Den Hekke                                              *)
(*                                                                                    *)
(*  This library is free software; you can redistribute it and/or                     *)
(*  modify it under the terms of the GNU Lesser General Public                        *)
(*  License as published by the Free Software Foundation; either                      *)
(*  version 2.1 of the License, or (at your option) any later version.                *)
(*                                                                                    *)
(*  This library is distributed in the hope that it will be useful,                   *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of                    *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU                 *)
(*  Lesser General Public License for more details.                                   *)
(*                                                                                    *)
(*  You should have received a copy of the GNU Lesser General Public                  *)
(*  License along with this library; if not, write to the Free Software               *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA         *)
(**************************************************************************************)

unit HexMemo;

interface

uses Windows, Messages, SysUtils, Classes, Controls, Graphics,
     Forms, Printers;

type
  THexMemo = class;

  TByteLines = class(TPersistent)
  private
    FOwner: THexMemo;
    FBuffer: array of Byte;
    FCapacity,
    FCount: Integer;
    function GetItems(I: Integer): Byte;
    procedure SetItems(I: Integer; const Value: Byte);
    procedure SetCount(const Value: Integer);
    procedure SetCapacity(const Value: Integer);
  protected
  public
    constructor Create(AOwner: THexMemo);
    destructor Destroy(); override;

    procedure LoadFromStream(St: TStream);
    procedure SaveToStream(St: TStream);

    procedure LoadFromFile(FileName: TFileName);
    procedure SaveToFile(FileName: TFileName);

    property Items[I: Integer]: Byte read GetItems write SetItems; default;
    property Count: Integer read FCount write SetCount;
    property Capacity: Integer read FCapacity write SetCapacity;
  end;

  THexMemo = class(TCustomControl)
  private
    FASCIIColor: TColor;
    FColor: TColor;
    FHexColor: TColor;
    FFont: TFont;
    FLines: TByteLines;
    FAddressColor: TColor;

    procedure DoScroll(Which, Action, Thumb: Integer);
    procedure WMHScroll(var M: TWMHScroll); message WM_HSCROLL;
    procedure WMVScroll(var M: TWMVScroll); message WM_VSCROLL;
    procedure WMSize(var M: TWMSize); message WM_SIZE;
    procedure WMGetDlgCode(var M: TWMGetDlgCode); message WM_GETDLGCODE;

    procedure SetASCIIColor(const Value: TColor);
    procedure SetHexColor(const Value: TColor);
    procedure SetFont(const Value: TFont);
    procedure SetAddressColor(const Value: TColor);
  protected
    FRange, FOrigin, FClientSize, FCharSize: TPoint;
    FOverhang, FPageSize, FPosition        : Integer;

    procedure Paint; override;
    procedure SetScrollBar;
    procedure KeyDown(var Key: Word; Shift: TShiftState); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure RecalcRange;
    procedure FontChanged(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ScrollTo(X, Y: Integer);

    property Bytes: TByteLines read FLines;
  published
    property Align;
    property AddressColor: TColor read FAddressColor write SetAddressColor default clWindowText;
    property ASCIIColor: TColor read FASCIIColor write SetASCIIColor default clWindowText;
    property Font: TFont read FFont write SetFont;
    property HexColor: TColor read FHexColor write SetHexColor default clWindowText;
    property ParentColor;
    property Color;
  end;

implementation

uses Math;

type
  TScrollKey = record
    sKey: Byte;
    Ctrl: Boolean;
    SBar: Byte;
    Action: Byte;
  end;

const
  cstBuffSize = 32768;
  cstLineMax  = 9 + (16 * 3) + 16;
  ScrollKeys: array[1..16] of TScrollKey = (
    (sKey: VK_LEFT;  Ctrl: False; SBar: SB_HORZ; Action: SB_LINEUP),
    (sKey: VK_RIGHT; Ctrl: False; SBar: SB_HORZ; Action: SB_LINEDOWN),
    (sKey: VK_LEFT;  Ctrl: True;  SBar: SB_HORZ; Action: SB_PAGEUP),
    (sKey: VK_RIGHT; Ctrl: True;  SBar: SB_HORZ; Action: SB_PAGEDOWN),
    (sKey: VK_HOME;  Ctrl: False; SBar: SB_HORZ; Action: SB_TOP),
    (sKey: VK_END;   Ctrl: False; SBar: SB_HORZ; Action: SB_BOTTOM),
    (sKey: VK_UP;    Ctrl: False; SBar: SB_VERT; Action: SB_LINEUP),
    (sKey: VK_DOWN;  Ctrl: False; SBar: SB_VERT; Action: SB_LINEDOWN),
    (sKey: VK_PRIOR; Ctrl: False; SBar: SB_VERT; Action: SB_PAGEUP),
    (sKey: VK_NEXT;  Ctrl: False; SBar: SB_VERT; Action: SB_PAGEDOWN),
    (sKey: VK_F1;    Ctrl: False; SBar: SB_VERT; Action: SB_PAGEDOWN),
    (sKey: VK_F2;    Ctrl: False; SBar: SB_VERT; Action: SB_PAGEUP),
    (sKey: VK_F3;    Ctrl: False; SBar: SB_VERT; Action: SB_TOP),
    (sKey: VK_F4;    Ctrl: False; SBar: SB_VERT; Action: SB_BOTTOM),
    (sKey: VK_HOME;  Ctrl: True;  SBar: SB_VERT; Action: SB_TOP),
    (sKey: VK_END;   Ctrl: True;  SBar: SB_VERT; Action: SB_BOTTOM)
  );

{ THexMemo }

constructor THexMemo.Create(AOwner: TComponent);
begin
  inherited;
  Width  := 320;
  Height := 200;
  ParentColor := False;
  FFont       := TFont.Create;
  with (FFont) do begin
    Name      := 'Courier New';
    Size      := 10;
    OnChange  := FontChanged;
  end;
  FColor      := clWindow;
  FASCIIColor := clWindowText;
  FHexColor   := clWindowText;
  FLines      := TByteLines.Create(Self);

  FontChanged(nil);

  Enabled     := True;
end;

destructor THexMemo.Destroy;
begin
  FLines.Free;
  inherited;
end;

procedure THexMemo.DoScroll(Which, Action, Thumb: Integer);
  function GetNewPos(Pos, Page, Range: Integer): Integer;
  begin
    case Action of
      SB_LINEUP        : Result := Pos - 1;
      SB_LINEDOWN      : Result := Pos + 1;
      SB_PAGEUP        : Result := Pos - Page;
      SB_PAGEDOWN      : Result := Pos + Page;
      SB_TOP           : Result := 0;
      SB_BOTTOM        : Result := Range;
      SB_THUMBPOSITION,
      SB_THUMBTRACK    : Result := Thumb;
    else
      Result := Pos;
    end;
  end;
var X, Y: Integer;
begin
  X := FOrigin.X;
  Y := FOrigin.Y;
  case Which of
    SB_HORZ : X := GetNewPos(X, FClientSize.X div 2, FRange.X);
    SB_VERT : Y := GetNewPos(Y, FClientSize.Y, FRange.Y);
  end;
  ScrollTo(X, Y);
end;

procedure THexMemo.FontChanged(Sender: TObject);
var DC: HDC;
    Save: THandle;
    Metrics: TTextMetric;
begin
  DC := GetDC(0);
  Save := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, Save);
  ReleaseDC(0, DC);

  with Metrics do begin
    FCharSize.X := tmAveCharWidth;
    FCharSize.Y := tmHeight + tmExternalLeading;
    FOverhang   := Max(tmOverhang, tmMaxCharWidth - tmAveCharWidth);
  end;

  RecalcRange;
  Invalidate;
end;

procedure THexMemo.KeyDown(var Key: Word; Shift: TShiftState);
var I: Integer;
begin
  inherited;
  if(Key <> 0) then begin
    for I := Low(ScrollKeys) to High(ScrollKeys) do
      with(ScrollKeys[I]) do
        if((sKey = Key) and (Ctrl = (Shift = [ssCtrl]))) then begin
          DoScroll(SBar, Action, 0);
        end;
  end;
end;

procedure THexMemo.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
begin
  SetFocus;
  inherited;
end;

procedure THexMemo.Paint;
var J  : Integer;
    Pos: TPoint;
    ChaineAddr,
    ChaineHexa,
    ChaineASCII: String;
begin
  inherited;
  SetViewportOrgEx(Canvas.Handle, -FOrigin.X * FCharSize.X, 0, @Pos);

  Canvas.Brush.Color := Color;
  Canvas.Font := FFont;

  Pos.Y := FOrigin.Y;

  while (Pos.Y < ((FLines.Count div 16) + 1)) and (Pos.Y < ((FOrigin.Y + FPageSize) + 1)) do begin
    Pos.X := 0;

    ChaineAddr  := IntToHex(Pos.Y, 8);
    ChaineHexa  := '';
    ChaineASCII := '';
    for J := Pos.Y * 16 to (Pos.Y * 16) + 15 do begin
      if(J < FLines.Count) then begin
        ChaineHexa := ChaineHexa + IntToHex(FLines[J], 2) + ' ';
        if(FLines[J] in [32..128]) then
          ChaineASCII := ChaineASCII + Chr(FLines[J])
        else
          ChaineASCII := ChaineASCII + '.';
      end else begin
        ChaineHexa := ChaineHexa + '   ';
      end;
    end;

    Canvas.Font.Color  := FAddressColor;
    Canvas.TextOut(Pos.X, FCharSize.Y * (Pos.Y - FOrigin.Y), ChaineAddr);
    Inc(Pos.X, Canvas.TextWidth(ChaineAddr + ' '));
    Canvas.Font.Color  := FHexColor;
    Canvas.TextOut(Pos.X, FCharSize.Y * (Pos.Y - FOrigin.Y), ChaineHexa);
    Inc(Pos.X, Canvas.TextWidth(ChaineHexa + ' '));
    Canvas.Font.Color  := FASCIIColor;
    Canvas.TextOut(Pos.X, FCharSize.Y * (Pos.Y - FOrigin.Y), ChaineASCII);

    Inc(Pos.Y);
  end;
end;

procedure THexMemo.RecalcRange;
begin
  if(HandleAllocated) then begin
    FClientSize.X := ClientWidth div FCharSize.X;
    FClientSize.Y := ClientHeight div FCharSize.Y;

    FPageSize     := FClientSize.Y;

    FRange.X      := Max(0, cstLineMax - FClientSize.X);
    FRange.Y      := Max(0, (FLines.Count div 16) + 1 - FClientSize.Y);
    ScrollTo(Min(FOrigin.X, FRange.X), Min(FOrigin.Y, FRange.Y));
    SetScrollBar;
  end;
end;

procedure THexMemo.ScrollTo(X, Y: Integer);
var R         : TRect;
    OldOrigine: TPoint;
begin
  X := Max(0, Min(X, FRange.X));
  Y := Max(0, Min(Y, FRange.Y));

  if (X <> FOrigin.X) or (Y <> FOrigin.Y) then begin
    OldOrigine := FOrigin;
    FOrigin    := Point(X, Y);
    if(HandleAllocated) then begin
      R := Parent.ClientRect;
      ScrollWindowEx(Handle, (OldOrigine.X - X) * FCharSize.X,
                             (OldOrigine.Y - Y) * FCharSize.Y,
                             nil, @R, 0, @R, 0);
      if(Y <> OldOrigine.Y) then
        SetScrollPos(Handle, SB_VERT, Y, True);
      if(X <> OldOrigine.X) then
        SetScrollPos(Handle, SB_HORZ, X, True);
      InvalidateRect(Handle, @R, True);
      Update;
    end;
  end;
end;

procedure THexMemo.SetAddressColor(const Value: TColor);
begin
  FAddressColor := Value;
end;

procedure THexMemo.SetASCIIColor(const Value: TColor);
begin
  FASCIIColor := Value;
end;

procedure THexMemo.SetFont(const Value: TFont);
begin
  FFont := Value;
end;

procedure THexMemo.SetHexColor(const Value: TColor);
begin
  FHexColor := Value;
end;

procedure THexMemo.SetScrollBar;
const InfoStruct: TScrollInfo = (fMask: SIF_DISABLENOSCROLL);
begin
  if(HandleAllocated) then begin
    if(FRange.X * FCharSize.X > FClientSize.X) then begin
      SetScrollRange(Handle, SB_HORZ, 0, Max(1, FRange.X), False);
      SetScrollPos(Handle, SB_HORZ, FOrigin.X, False);
    end else
      SetScrollInfo(Handle, SB_HORZ, InfoStruct, False);

    if(FRange.Y * FCharSize.Y > FClientSize.Y) then begin
      SetScrollRange(Handle, SB_VERT, 0, Max(1, FRange.Y), False);
      SetScrollPos(Handle, SB_VERT, FOrigin.Y, False);
    end else
      SetScrollInfo(Handle, SB_VERT, InfoStruct, False);
  end;
end;

procedure THexMemo.WMGetDlgCode(var M: TWMGetDlgCode);
begin
  M.Result := DLGC_WANTARROWS or DLGC_WANTCHARS;
end;

procedure THexMemo.WMHScroll(var M: TWMHScroll);
begin
  DoScroll(SB_HORZ, M.ScrollCode, M.Pos);
end;

procedure THexMemo.WMSize(var M: TWMSize);
begin
  inherited;
  RecalcRange;
end;

procedure THexMemo.WMVScroll(var M: TWMVScroll);
begin
  DoScroll(SB_VERT, M.ScrollCode, M.Pos);
end;

{ TByteLines }

constructor TByteLines.Create(AOwner: THexMemo);
begin
  FOwner := AOwner;
  FCapacity := 0;
  FCount := 0;
end;

destructor TByteLines.Destroy;
begin
  Capacity := 0;
  inherited;
end;

function TByteLines.GetItems(I: Integer): Byte;
begin
  result := FBuffer[I];
end;

procedure TByteLines.LoadFromFile(FileName: TFileName);
var St: TFileStream;
begin
  St := TFileStream.Create(FileName, fmOpenRead or fmShareDenyWrite);
  try
    LoadFromStream(St);
  finally
    St.Free;
  end;
end;

procedure TByteLines.LoadFromStream(St: TStream);
begin
  Capacity := St.Size;
  Count    := Capacity;

  while(St.Position < St.Size) do
    St.Read(FBuffer[St.Position], Min(St.Size - St.Position, cstBuffSize));

  FOwner.RecalcRange;
  FOwner.Invalidate;
end;

procedure TByteLines.SaveToFile(FileName: TFileName);
begin

end;

procedure TByteLines.SaveToStream(St: TStream);
begin

end;

procedure TByteLines.SetCapacity(const Value: Integer);
begin
  if(Value <> FCapacity) then begin
    SetLength(FBuffer, Value);
    FCapacity := Value;
  end;
end;

procedure TByteLines.SetCount(const Value: Integer);
begin
  if(Value <> FCount) then begin
    if(Value > Capacity) then begin

      if(Capacity < 16) then
        Capacity := Value + 4
      else if(Capacity < 64) then
        Capacity := Value + 8
      else
        Capacity := Value + (Value div 8);

      FillChar(FBuffer[FCount], Value - FCount, #0);
    end;

    FCount := Value;
  end;
  FOwner.RecalcRange;
  FOwner.Invalidate;
end;

procedure TByteLines.SetItems(I: Integer; const Value: Byte);
begin
  if(I > Count) then
    Count := I + 1;

  FBuffer[I] := Value;
  FOwner.RecalcRange;
  FOwner.Invalidate;
end;

end.
