(**************************************************************************************)
(*  'tils Lib - Librairie de composants divers                                        *)
(*  Copyright (C) 2003  U. Van Den Hekke                                              *)
(*                                                                                    *)
(*  This library is free software; you can redistribute it and/or                     *)
(*  modify it under the terms of the GNU Lesser General Public                        *)
(*  License as published by the Free Software Foundation; either                      *)
(*  version 2.1 of the License, or (at your option) any later version.                *)
(*                                                                                    *)
(*  This library is distributed in the hope that it will be useful,                   *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of                    *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU                 *)
(*  Lesser General Public License for more details.                                   *)
(*                                                                                    *)
(*  You should have received a copy of the GNU Lesser General Public                  *)
(*  License along with this library; if not, write to the Free Software               *)
(*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA         *)
(**************************************************************************************)

unit VStream;

interface

uses Windows, SysUtils, Classes;

type
  TOnNewStream    = function(var NewStream: TStream; var NewSpace: LongInt): boolean of object;
  TBufferedStream = class(TStream)
  private
    FStream       : TStream;
    tempFile      : string;

    function GetMemory(): Pointer;
  protected
  public
    constructor Create();
    destructor Destroy; override;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;

    property Memory : Pointer read GetMemory;
  end;

  TMultipleStream = class(TStream)
  private
    FCurrentStream: TStream;
    FMaxSpace     : LongInt;
    FCRC32        : LongWord;
    FOnNewStream  : TOnNewStream;
    FBytesWritten : LongInt;
    procedure SetOnNewStream(const Value: TOnNewStream);
    procedure SetBytesWritten(const Value: LongInt);
    function GetCRC32: LongWord;
  protected
  public
    constructor Create(const Stream: TStream; const Space: LongInt; const OnNew: TOnNewStream);
    destructor Destroy(); override;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;

    property OnNewStream : TOnNewStream read FOnNewStream write SetOnNewStream;
    property CRC32       : LongWord read GetCRC32;
    property BytesWritten: LongInt read FBytesWritten write SetBytesWritten;
  end;

  ELimitedStream = class(EStreamError);
  TLimitedStream = class(TStream)
  private
    FStream: TStream;
    FPSize,
    FPosition,
    FSize     : Int64;
    FModified : boolean;
  protected
  public
    constructor Create(const Stream: TStream);
    destructor Destroy(); override;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

function MoveStreamData(const St: TStream; const PosSource, PosDest, Count: Integer): integer;

implementation

uses FileTools, Math, Hash;

function MoveStreamData(const St: TStream; const PosSource, PosDest, Count: Integer): integer;
const BuffSize = 1024;
var   Buffer   : array[1..BuffSize] of byte;
      CntRW    : integer;
begin
  result := 0;
  if(PosSource = PosDest) then Exit;
  if(PosSource > PosDest) then begin
    while(result < Count) do begin
      St.Seek(PosSource + result, soFromBeginning);
      if(Count - Result < BuffSize) then
        CntRW := St.Read(Buffer, Count - Result)
      else
        CntRW := St.Read(Buffer, BuffSize);

      St.Seek(PosDest + result, soFromBeginning);
      CntRW := St.Write(Buffer, CntRW);

      Inc(result, CntRW);
    end;
  end else begin
    while(result < Count) do begin
      CntRW := Min(BuffSize, Count - Result);
      St.Seek(((PosSource + Count) - Result) - CntRW, soFromBeginning);
      CntRW := St.Read(Buffer, CntRW);

      St.Seek(((PosDest + Count) - Result) - CntRW, soFromBeginning);
      CntRW := St.Write(Buffer, CntRW);

      Inc(result, CntRW);
    end;
  end;
end;

{ TBufferedStream }

constructor TBufferedStream.Create();
begin
  inherited Create();
  tempFile      := GenerateFileName();
  FStream       := TFileStream.Create(tempFile, fmCreate);
end;

destructor TBufferedStream.Destroy;
begin
  SysUtils.DeleteFile(tempFile);
  FStream.Free;
  inherited Destroy();
end;

function TBufferedStream.Read(var Buffer; Count: Longint): Longint;
begin
  result := FStream.read(Buffer, Count);
end;

function TBufferedStream.Write(const Buffer; Count: Longint): Longint;
begin
  result := FStream.write(Buffer, Count);
end;

function TBufferedStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  result := FStream.Seek(Offset, Origin);
end;

function TBufferedStream.GetMemory(): Pointer;
var Mem: TMemoryStream;
begin
  if(FStream is TMemoryStream) then
    result := (FStream as TMemoryStream).Memory
  else begin
    Mem := TMemoryStream.Create();
    Mem.CopyFrom(FStream, FStream.Size);
    result := Mem.Memory;
  end;
end;

{ TMultipleStream }

constructor TMultipleStream.Create(const Stream: TStream;
  const Space: LongInt; const OnNew: TOnNewStream);
begin
  inherited Create();
  FCurrentStream := Stream;
  FMaxSpace      := Space;
  FBytesWritten  := 0;
  FCRC32         := $FFFFFFFF;
  FOnNewStream   := OnNew;
end;

destructor TMultipleStream.Destroy;
begin
  inherited Destroy();
end;

function TMultipleStream.GetCRC32: LongWord;
begin
  result := not FCRC32;
end;

function TMultipleStream.Read(var Buffer; Count: Integer): Longint;
var Buff, S: Pointer;
begin
  GetMem(Buff, Count);
  try
    result := FCurrentStream.Read(Buff^, Count);
    if(result < Count) then begin
      if(FOnNewStream(FCurrentStream, FMaxSpace)) then begin;
        S := Pointer(Integer(Buff) + result);
        result := result + FCurrentStream.Read(S^, Count - result);
      end;
    end;
    Move(Buff^, Buffer, Count);
  finally
    FreeMem(Buff);
  end;
  FCRC32 := Hash.CRC32(FCRC32, @Buffer, Result);
end;

function TMultipleStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  result := FCurrentStream.Seek(Offset, Origin);
end;

procedure TMultipleStream.SetBytesWritten(const Value: LongInt);
begin
  FBytesWritten := Value;
end;

procedure TMultipleStream.SetOnNewStream(const Value: TOnNewStream);
begin
  FOnNewStream := Value;
end;

function TMultipleStream.Write(const Buffer; Count: Integer): Longint;
var Delta  : Integer;
    Buff, S: Pointer;
begin
  if((LongWord(FCurrentStream.Size) + LongWord(Count)) <= LongWord(FMaxSpace)) then
    result := FCurrentStream.Write(Buffer, Count)
  else begin
    Delta  := Integer(FMaxSpace) - FCurrentStream.Size;
    if(Delta > 0) then
      result := FCurrentStream.Write(Buffer, Delta)
    else
      result := 0;

    FOnNewStream(FCurrentStream, FMaxSpace);

    GetMem(Buff, Count);
    try
      System.Move(Buffer, Buff^, Count);
      S := Pointer(Integer(Buff) + Delta);
      result := result + FCurrentStream.Write(S^, Count - Delta);
    finally
      FreeMem(Buff);
    end;
  end;
  Inc(FBytesWritten, Result);
  FCRC32 := Hash.CRC32(FCRC32, @Buffer, Result);
end;

{ TLimitedStream }

constructor TLimitedStream.Create(const Stream: TStream);
begin
  inherited Create;
  FStream := Stream;
  FPSize  := FStream.Position;
  if(FStream.Read(FSize, SizeOf(Integer)) < SizeOf(Integer)) then begin
    FSize := 0;
    FStream.Write(FSize, SizeOf(Integer))
  end;
  FPosition := 0;
  FModified := false;
end;

destructor TLimitedStream.Destroy;
var Pos: Int64;
begin
  if(FModified) then begin
    Pos := FStream.Position;
    FStream.Seek(FPSize, soBeginning);
    FStream.Write(FSize, SizeOf(Integer));
    FStream.Seek(Pos, soBeginning);
  end;
  inherited;
end;

function TLimitedStream.Read(var Buffer; Count: Integer): Longint;
begin
  Inc(FPosition, Count);
  if(FPosition > FSize) then begin
    result := FStream.Read(Buffer, Count - (FPosition - FSize));
    FPosition := FSize;
  end else
    result := FStream.Read(Buffer, Count);
end;

function TLimitedStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  case(Origin) of
    soFromBeginning : begin
      FStream.Seek(-FPosition + Offset, soCurrent);
      FPosition := Offset;
    end;
    soFromCurrent : begin
      Inc(FPosition, Offset);
      if(FPosition > FSize) then begin
        FStream.Seek(Offset - (FSize - FPosition), soCurrent);
        FPosition := FSize;
      end else
        FStream.Seek(Offset, soCurrent);
    end;
    soFromEnd : begin
      FStream.Seek(FSize - FPosition + Offset, soCurrent);
      FPosition := FSize + Offset;
    end;
  end;
  result := FPosition;
end;

function TLimitedStream.Write(const Buffer; Count: Integer): Longint;
begin
  Inc(FPosition, Count);
  if(FPosition > FSize) then
    FSize := FPosition;
  result := FStream.Write(Buffer, Count);
  FModified := true;
end;

end.


